@@ -3,7 +3,7 @@
  * evemu - Kernel device emulation
  *
  * Copyright (C) 2010-2012 Canonical Ltd.
- * Copyright (C) 2013 Red Hat, Inc.
+ * Copyright (C) 2013-2015 Red Hat, Inc.
  *
  * This library is free software: you can redistribute it and/or modify it
  * under the terms of the GNU Lesser General Public License version 3
@@ -52,13 +52,14 @@
 #include <poll.h>
 #include <ctype.h>
 #include <unistd.h>
+#include <sys/utsname.h>
 
 #include "version.h"
 
 /* File format version we write out
    NOTE: if you bump the version number, make sure you update README */
 #define EVEMU_FILE_MAJOR 1
-#define EVEMU_FILE_MINOR 2
+#define EVEMU_FILE_MINOR 3
 
 #define SYSCALL(call) while (((call) == -1) && (errno == EINTR))
 
@@ -68,6 +69,9 @@ enum error_level {
 	FATAL,
 };
 
+static int error(enum error_level level, const char *format, ...)
+	        __attribute__ ((format (printf, 2, 3)));
+
 static int error(enum error_level level, const char *format, ...)
 {
 	va_list args;
@@ -134,6 +138,9 @@ struct evemu_device *evemu_new(const char *name)
 
 void evemu_delete(struct evemu_device *dev)
 {
+	if (dev == NULL)
+		return;
+
 	if (dev->uidev)
 		evemu_destroy(dev);
 	libevdev_free(dev->evdev);
@@ -247,12 +254,12 @@ static void write_prop(FILE * fp, const struct evemu_device *dev)
 	int i;
 	unsigned char mask[max(8, (INPUT_PROP_MAX + 7)/8)] = {0};
 
-	for (i = 0; i < INPUT_PROP_MAX; i ++) {
+	for (i = 0; i < INPUT_PROP_CNT; i ++) {
 		if (evemu_has_prop(dev, i))
 			set_bit(mask, i);
 	}
 
-	for (i = 0; i < (INPUT_PROP_MAX + 7)/8; i +=8) {
+	for (i = 0; i < (INPUT_PROP_CNT + 7)/8; i +=8) {
 		fprintf(fp, "P: %02x %02x %02x %02x %02x %02x %02x %02x\n",
 			mask[i], mask[i + 1], mask[i + 2], mask[i + 3],
 			mask[i + 4], mask[i + 5], mask[i + 6], mask[i + 7]);
@@ -277,11 +284,11 @@ static void write_mask(FILE * fp, const struct evemu_device *dev)
 		if (max == -1)
 			continue;
 
-		for (code = 0; code < (unsigned int)max; code++)
+		for (code = 0; code <= (unsigned int)max; code++)
 			if (evemu_has_event(dev, type, code))
 				set_bit(mask, code);
 
-		for (i = 0; i < (max + 7)/8; i += 8) {
+		for (i = 0; i < ((max + 1) + 7)/8; i += 8) {
 			fprintf(fp, "B: %02x %02x %02x %02x %02x %02x %02x %02x %02x\n",
 				type, mask[i], mask[i + 1], mask[i + 2], mask[i + 3],
 				mask[i + 4], mask[i + 5], mask[i + 6], mask[i + 7]);
@@ -295,46 +302,89 @@ static void write_abs(FILE *fp, int index, const struct input_absinfo *abs)
 		abs->minimum, abs->maximum, abs->fuzz, abs->flat, abs->resolution);
 }
 
+static void write_led(FILE *fp, int index, int state)
+{
+	fprintf(fp, "L: %02x %d\n", index, state);
+}
+
+static void write_sw(FILE *fp, int index, int state)
+{
+	fprintf(fp, "S: %02x %d\n", index, state);
+}
+
 /* Print an evtest-like description */
 static void write_desc(const struct evemu_device *dev, FILE *fp)
 {
 	int i, j;
+	int state;
 	fprintf(fp, "# Input device name: \"%s\"\n", evemu_get_name(dev));
 	fprintf(fp, "# Input device ID: bus %#04x vendor %#04x product %#04x version %#04x\n",
 		evemu_get_id_bustype(dev), evemu_get_id_vendor(dev),
 		evemu_get_id_product(dev), evemu_get_id_version(dev));
+
+	if (evemu_has_event(dev, EV_ABS, ABS_X) &&
+	    evemu_has_event(dev, EV_ABS, ABS_Y)) {
+		int min, max, res;
+		int w = 0, h = 0;
+
+		min = evemu_get_abs_minimum(dev, ABS_X);
+		max = evemu_get_abs_maximum(dev, ABS_X);
+		res = evemu_get_abs_resolution(dev, ABS_X);
+		if (res != 0)
+			w = (max - min)/res;
+
+		min = evemu_get_abs_minimum(dev, ABS_Y);
+		max = evemu_get_abs_maximum(dev, ABS_Y);
+		res = evemu_get_abs_resolution(dev, ABS_Y);
+		if (res != 0)
+			h = (max - min)/res;
+
+		if (w != 0 && h != 0)
+			fprintf(fp, "# Size in mm: %dx%d\n", w, h);
+	}
+
 	fprintf(fp, "# Supported events:\n");
-	for (i = 0; i < EV_MAX; i++) {
+	for (i = 0; i < EV_CNT; i++) {
 		if (!evemu_has_bit(dev, i))
 			continue;
 
 		fprintf(fp, "#   Event type %d (%s)\n", i, libevdev_event_type_get_name(i));
-		for (j = 0; j < libevdev_event_type_get_max(i); j++) {
+		for (j = 0; j <= libevdev_event_type_get_max(i); j++) {
 			if (!evemu_has_event(dev, i, j))
 				continue;
 
 			fprintf(fp, "#     Event code %d (%s)\n",
 				    j, libevdev_event_code_get_name(i, j));
-			if (i == EV_ABS) {
-				fprintf(fp, "#       Value %6d\n"
-					    "#       Min   %6d\n"
-					    "#       Max   %6d\n"
-					    "#       Fuzz  %6d\n"
-					    "#       Flat  %6d\n"
-					    "#       Resolution %d\n",
-					    evemu_get_abs_current_value(dev, j),
-					    evemu_get_abs_minimum(dev, j),
-					    evemu_get_abs_maximum(dev,j),
-					    evemu_get_abs_fuzz(dev, j),
-					    evemu_get_abs_flat(dev, j),
-					    evemu_get_abs_resolution(dev, j));
+			switch(i) {
+				case EV_ABS:
+					fprintf(fp,
+						"#       Value   %6d\n"
+						"#       Min     %6d\n"
+						"#       Max     %6d\n"
+						"#       Fuzz    %6d\n"
+						"#       Flat    %6d\n"
+						"#       Resolution %3d\n",
+						evemu_get_abs_current_value(dev, j),
+						evemu_get_abs_minimum(dev, j),
+						evemu_get_abs_maximum(dev,j),
+						evemu_get_abs_fuzz(dev, j),
+						evemu_get_abs_flat(dev, j),
+						evemu_get_abs_resolution(dev, j));
+					break;
+				case EV_LED:
+				case EV_SW:
+					state = libevdev_get_event_value(dev->evdev, i, j);
+					fprintf(fp, "#        State %d\n", state);
+					break;
+				default:
+					break;
 			}
 		}
 	}
 
 #ifdef INPUT_PROP_MAX
 	fprintf(fp, "# Properties:\n");
-	for (i = 0; i < INPUT_PROP_MAX; i++) {
+	for (i = 0; i < INPUT_PROP_CNT; i++) {
 		if (!evemu_has_prop(dev, i))
 			continue;
 		fprintf(fp, "#   Property  type %d (%s)\n", i,
@@ -343,11 +393,35 @@ static void write_desc(const struct evemu_device *dev, FILE *fp)
 #endif
 }
 
+static void
+write_header(FILE *fp)
+{
+	struct utsname u;
+	char modalias[2048];
+	FILE *dmi;
+
+	fprintf(fp, "# EVEMU %d.%d\n", EVEMU_FILE_MAJOR, EVEMU_FILE_MINOR);
+
+	if (uname(&u) == -1)
+		return;
+
+	fprintf(fp, "# Kernel: %s\n", u.release);
+
+	dmi = fopen("/sys/class/dmi/id/modalias", "r");
+	if (dmi) {
+		if (fgets(modalias, sizeof(modalias), dmi)) {
+			fprintf(fp, "# DMI: %s", modalias);
+		}
+		fclose(dmi);
+	}
+}
+
 int evemu_write(const struct evemu_device *dev, FILE *fp)
 {
 	int i;
+	int state;
 
-	fprintf(fp, "# EVEMU %d.%d\n", EVEMU_FILE_MAJOR, EVEMU_FILE_MINOR);
+	write_header(fp);
 
 	write_desc(dev, fp);
 
@@ -366,6 +440,16 @@ int evemu_write(const struct evemu_device *dev, FILE *fp)
 		if (evemu_has_event(dev, EV_ABS, i))
 			write_abs(fp, i, libevdev_get_abs_info(dev->evdev, i));
 
+	for (i = 0; i < LED_CNT; i++)
+		if (evemu_has_event(dev, EV_LED, i) &&
+		    (state = libevdev_get_event_value(dev->evdev, EV_LED, i)) != 0)
+			write_led(fp, i, state);
+
+	for (i = 0; i < SW_CNT; i++)
+		if (evemu_has_event(dev, EV_SW, i) &&
+		    (state = libevdev_get_event_value(dev->evdev, EV_SW, i)) != 0)
+			write_sw(fp, i, state);
+
 	return 0;
 }
 
@@ -374,7 +458,7 @@ static int parse_name(struct evemu_device *dev, const char *line)
 	int matched;
 	char *devname = NULL;
 
-	if ((matched = sscanf(line, "N: %m[^\n]\n", &devname)) > 0) {
+	if ((matched = sscanf(line, "N: %m[^\n\r]\n", &devname)) > 0) {
 		if (strlen(evemu_get_name(dev)) == 0)
 			evemu_set_name(dev, devname);
 	}
@@ -449,7 +533,12 @@ static int parse_mask(struct evemu_device *dev, const char *line)
 				mask + 4, mask + 5, mask + 6, mask + 7);
 
 	if (matched != 9) {
-		error(WARNING, "Invalid EV_BIT line. Parsed %d numbers, expected 9: %s", matched, line);
+		error(FATAL, "Invalid EV_BIT line. Parsed %d numbers, expected 9: %s", matched, line);
+		return -1;
+	}
+
+	if (index >= EV_CNT) {
+		error(FATAL, "Invalid EV_* index %#x in line: %s", index, line);
 		return -1;
 	}
 
@@ -501,6 +590,50 @@ static int parse_abs(struct evemu_device *dev, const char *line, struct version
 	return 1;
 }
 
+static int parse_led(const char *line)
+{
+	int matched;
+	unsigned int index;
+	int state;
+
+	if (strlen(line) <= 2 || strncmp(line, "L:", 2) != 0)
+		return 0;
+
+	matched = sscanf(line, "L: %02x %d\n", &index, &state);
+
+	if (matched != 2) {
+		error(FATAL, "Invalid EV_LED line. Parsed %d numbers, expected 2: %s", matched, line);
+		return -1;
+	}
+
+	/* We can't set the LEDs directly, we'd have to send an event
+	 * through the device but that's potentially racy */
+
+	return 1;
+}
+
+static int parse_sw(const char *line)
+{
+	int matched;
+	unsigned int index;
+	int state;
+
+	if (strlen(line) <= 2 || strncmp(line, "S:", 2) != 0)
+		return 0;
+
+	matched = sscanf(line, "S: %02x %d\n", &index, &state);
+
+	if (matched != 2) {
+		error(FATAL, "Invalid EV_SW line. Parsed %d numbers, expected 2: %s", matched, line);
+		return -1;
+	}
+
+	/* We can't set the switches directly, we'd have to send an event
+	 * through the device but that's potentially racy */
+
+	return 1;
+}
+
 static struct version parse_file_format_version(const char *line)
 {
 	struct version v;
@@ -579,26 +712,52 @@ int evemu_read(struct evemu_device *dev, FILE *fp)
 	if (rc == -1)
 		goto out;
 
+	while((rc = parse_led(line)) > 0)
+		if (!next_line(fp, &line, &size))
+			break;
+	if (rc == -1)
+		goto out;
+
+	while((rc = parse_sw(line)) > 0)
+		if (!next_line(fp, &line, &size))
+			break;
+	if (rc == -1)
+		goto out;
+
 	rc = 1;
 
+	fseek(fp, -strlen(line), SEEK_CUR);
+
 out:
 	free(line);
 	return rc;
 }
 
+static inline unsigned long millis(const struct timeval *tv)
+{
+	return tv->tv_sec * 1000 + tv->tv_usec/1000;
+}
+
 static int write_event_desc(FILE *fp, const struct input_event *ev)
 {
 	int rc;
+	static unsigned long last_ms = 0;
+	unsigned long time, dt;
 
 	if (ev->type == EV_SYN) {
-		if (ev->code == SYN_MT_REPORT)
+		if (ev->code == SYN_MT_REPORT) {
 			rc = fprintf(fp, "# ++++++++++++ %s (%d) ++++++++++\n",
 				     libevdev_event_code_get_name(ev->type, ev->code),
 				     ev->value);
-		else
-			rc = fprintf(fp, "# ------------ %s (%d) ----------\n",
+		} else {
+			time = millis(&ev->time);
+			dt = time - last_ms;
+			last_ms = time;
+			rc = fprintf(fp, "# ------------ %s (%d) ---------- +%ldms\n",
 				     libevdev_event_code_get_name(ev->type, ev->code),
-				     ev->value);
+				     ev->value,
+				     dt);
+		}
 	} else {
 		rc = fprintf(fp, "# %s / %-20s %d\n",
 			     libevdev_event_type_get_name(ev->type),
@@ -619,13 +778,13 @@ int evemu_write_event(FILE *fp, const struct input_event *ev)
 }
 
 static inline long time_to_long(const struct timeval *tv) {
-	return tv->tv_sec * 1000000 + tv->tv_usec;
+	return tv->tv_sec * 1000000L + tv->tv_usec;
 }
 
 static inline struct timeval long_to_time(long time) {
 	struct timeval tv;
-	tv.tv_sec = time/1000000;
-	tv.tv_usec = time % 1000000;
+	tv.tv_sec = time/1000000L;
+	tv.tv_usec = time % 1000000L;
 	return tv;
 }
 
@@ -644,7 +803,7 @@ int evemu_record(FILE *fp, int fd, int ms)
 			long time;
 
 			if (offset == 0)
-				offset = time_to_long(&ev.time);
+				offset = time_to_long(&ev.time) - 1;
 
 			time = time_to_long(&ev.time);
 			ev.time = long_to_time(time - offset);
@@ -701,10 +860,21 @@ int evemu_create_event(struct input_event *ev, int type, int code, int value)
 	return 0;
 }
 
+static inline unsigned long s2us(unsigned long s)
+{
+	return s * 1000000L;
+}
+
+static inline unsigned long us2s(unsigned long us)
+{
+	return us / 1000000L;
+}
+
 int evemu_read_event_realtime(FILE *fp, struct input_event *ev,
 			      struct timeval *evtime)
 {
 	unsigned long usec;
+	const unsigned long ERROR_MARGIN = 150; /* µs */
 	int ret;
 
 	ret = evemu_read_event(fp, ev);
@@ -712,12 +882,13 @@ int evemu_read_event_realtime(FILE *fp, struct input_event *ev,
 		return ret;
 
 	if (evtime) {
-		if (!evtime->tv_sec)
+		if (evtime->tv_sec == 0 && evtime->tv_usec == 0)
 			*evtime = ev->time;
-		usec = 1000000L * (ev->time.tv_sec - evtime->tv_sec);
-		usec += ev->time.tv_usec - evtime->tv_usec;
-		if (usec > 500) {
-			usleep(usec);
+		usec = time_to_long(&ev->time) - time_to_long(evtime);
+		if (usec > ERROR_MARGIN * 2) {
+			if (usec > s2us(10))
+				error(INFO, "Sleeping for %lds.\n", us2s(usec));
+			usleep(usec - ERROR_MARGIN);
 			*evtime = ev->time;
 		}
 	}
